
package mobicomp.hearts.communication;


import java.io.IOException;
import java.util.Iterator;

import mobicomp.hearts.data.Player;
import mobicomp.hearts.lobby.Game;
import mobicomp.hearts.lobby.LobbyData;
import mobicomp.interfaces.MHMessage;
import mobicomp.interfaces.MHSocket;
import mobicomp.net.MHMessageImpl;
import mobicomp.net.MHSocketImpl;
import mobicomp.net.MessageListener;


/**
 * @author Andreas Wetzel <andi@spamato.net>
 * last update 7.1.2005 Nicolas
 */
public class HeartsCommunication implements MessageListener
{
    private static final int MAX_SEARCH_DISTANCE = 4;

    private static final int MAX_PLAYERS = 4;
    
    
    
    private LobbyData data;
    private Thread searchThread;
    private MHSocket layer;

    public HeartsCommunication(LobbyData data) {
        this.data = data;
        try {
			layer = new MHSocketImpl(data.player.getId(), this);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
    }
    
    /**
     * send a MHMessage using the multi-hop-layer 
     * @param mhm The message to be sent
     */
    private void send( MHMessage mhm) {
        try {
            layer.send( mhm);
        }
        catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    /**
     * This method is called if the user wants to look for open games
     * It starts a new Thread and floods the network with a LOOK_FOR_GAMES message
     */
    public void startLookingForOpenGames() {
        System.out.println( data.player.getId() + ": starting search thread");
        searchThread = new Thread( "searchThread")
        {
            public void run() {
                for (int i = 1; i <= MAX_SEARCH_DISTANCE; i++) {
                    if ( !data.isLookingForGames ) {
                        continue;
                    }
                    HeartsMessage message = new HeartsMessage( HeartsMessage.LOOK_FOR_GAMES, null);
        
                    // initiate the MHMessage
                    MHMessage mhm = new MHMessageImpl();        
            		mhm.setType(MHMessage.TYPE_SRMSG);
            		mhm.setSender(data.player.getId());
            		mhm.setReceiver(MHSocket.BROADCAST_ADDRESS);
            		mhm.setData(message.toByteArray());
                    
                    mhm.setTTL( i);
                    send( mhm);
                    if ( i < MAX_SEARCH_DISTANCE ) {
                        try {
                            Thread.sleep( 5000);
                        }
                        catch (InterruptedException e) {
                            // TODO Auto-generated catch block
                            e.printStackTrace();
                        }
                    }
                }

                // set lookingForGames back to false
                data.isLookingForGames = false;
                data.notifyObservers();
                System.out.println( data.player.getId() + ": search thread ended");
            }
        };
        searchThread.start();
    }

    /**
     * This method is called to join a specific remote game hosted by 
     * remotePlayer
     * @param remotePlayer Player hosting the remote game
     */
    public void joinRemoteGame( Player remotePlayer) {
        System.out.println( data.player.getId() + ": joining remote game hosted by " + remotePlayer.getName());
        HeartsMessage message = new HeartsMessage( HeartsMessage.JOIN_GAME_REQ, data.player);
           
        // create a new Multihop message containing the LobbyMessage
        MHMessage mhm = new MHMessageImpl();        
		mhm.setType(MHMessage.TYPE_SRMSG);
		mhm.setSender(data.player.getId());
		mhm.setReceiver(remotePlayer.getId());
		mhm.setData(message.toByteArray());
        
        send( mhm);
    }

    /**
     * Used to leave a joined but not yet started remote game
     * @param remotePlayer Player hosting the game 
     */
    public void leaveRemoteGame( Player remotePlayer) {
        System.out.println( data.player.getId() + ": leaving remote game hosted by " + remotePlayer.getName());
        HeartsMessage message = new HeartsMessage( HeartsMessage.LEAVE_GAME, data.player);
        
        // create a new Multihop message containing the LobbyMessage
        MHMessage mhm = new MHMessageImpl();        
		mhm.setType(MHMessage.TYPE_SRMSG);
		mhm.setSender(data.player.getId());
		mhm.setReceiver(remotePlayer.getId());
		mhm.setData(message.toByteArray());
		send( mhm);
    }

    /**
     * This method is called to add a remote player to a locally hosted game
     * @param remotePlayer The player interested in joining the game
     */
    public void addRemotePlayer( Player remotePlayer) {
        System.out.println( data.player.getId() + ": " + remotePlayer.getName() + " tries to join my game");
        if ( data.isHostingGame && data.currentGame.addPlayer(remotePlayer)){ // if there is a locally hosted game and the new player was added correctly
                	
           	HeartsMessage message = new HeartsMessage( HeartsMessage.JOIN_GAME_ACK, null);
            // send a JOIN_ACK To the remote player
            MHMessage mhm = new MHMessageImpl();        
    		mhm.setType(MHMessage.TYPE_SRMSG);
    		mhm.setSender(data.player.getId());
    		mhm.setReceiver(remotePlayer.getId());
    		mhm.setData(message.toByteArray());
    		
    		// HACK: There seems to be a bug leading to a problem 
    		// if this message directly follows a MHAckMsg
    		// Needs to be fixed
    		try {
				Thread.sleep(500);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
    		
    		send( mhm);
        	
    		// notify the gui og the changes
    		data.notifyObservers();
    		// notify the other joined players
    		notifyRegisteredPlayers(HeartsMessage.HAVE_GAME);
              
         }
    }

    /**
     * This method is used to remove a remote player from a locally hosted game
     * as a reaction on receiving a LEAVE_GAME message
     * @param remotePlayer The player to remove from the game
     */
    public void removeRemotePlayer( Player remotePlayer) {
        System.out.println( data.player.getId() + ": " + remotePlayer.getId() + " left my game");

        if ( data.isHostingGame ) {
            data.currentGame.removePlayer( remotePlayer);

            // notify remote player
            announceGame(remotePlayer.getId(), HeartsMessage.HAVE_GAME);

            data.notifyObservers();
            notifyRegisteredPlayers(HeartsMessage.HAVE_GAME);
        }
    }
    
    /**
     * Called to abort a locally hosted but not yet started game
     *
     */
    public void abortLocaldGame(){
    	System.out.println( data.player.getId() + ": Abort local game");
    	notifyRegisteredPlayers(HeartsMessage.ABORT_GAME);	 
    }

   /**
    * Used to send a specific command to all joined players
    * The message also contains the Game Object of this game as payload
    * @param command Type of the sent HeartsMessage 
    */
    private void notifyRegisteredPlayers(int command) {
        Iterator it = data.currentGame.getPlayers().iterator(); 
        while(it.hasNext()) {
            Player p = (Player) it.next();
            // if not local player, send update...
            if ( !p.equals( data.player) ) {
                announceGame(p.getId(), command);
            }
        }
    }

    /**
     * Helper function to send an update of the game state to a specific player
     * @param remoteId Remote node to inform	
     * @param command	Type of HeartsMesage to send
     */
    private void announceGame(int remoteId, int command) {
        HeartsMessage reply = new HeartsMessage(command, data.currentGame);
        MHMessage mhm = new MHMessageImpl();
        
		mhm.setType(MHMessage.TYPE_SRMSG);
		mhm.setSender(data.player.getId());
		mhm.setReceiver(remoteId);
		mhm.setData(reply.toByteArray());
        
        send(mhm);
    }

    /**
     * This method is called before starting a local game. A corresponding START_GAME 
     * message is sent to all joined remote players. Then the actual game is started
     * locally.
     * @param game The game object of the starting game
     */
    public void startGame(Game game) {
    	Iterator it = game.getPlayers().iterator();

        while (it.hasNext()) {
            Player p = (Player) it.next();
            if ( !p.equals( game.getHostPlayer()) ) {
                HeartsMessage lm = new HeartsMessage(HeartsMessage.START_GAME, game);
                MHMessage mhm = new MHMessageImpl();
        		mhm.setType(MHMessage.TYPE_SRMSG);
        		mhm.setSender(data.player.getId());
        		mhm.setReceiver(p.getId());
        		mhm.setData(lm.toByteArray());
        		// send a start message to all players
                send(mhm);
            }
        }
        // start the game locally
        startGameThread(game);
    }

    /**
     * TO COME SOON...
     * @param game
     */
    private void startGameThread( Game game) {
    	System.out.println("Starting game");

    }

    /**
     * @see ch.ethz.dcg.mobicomp.net.MessageListener#onMessage(ch.ethz.dcg.mobicomp.net.Message)
     */
    public void onMessage(MHMessage message) {
        HeartsMessage m = HeartsMessage.parseBytes(message.getData());
        switch (m.getType()) {
            case HeartsMessage.LOOK_FOR_GAMES:
                // if i am hosting, i have to answer to LOOK_FOR_GAMES packets if my game is not full...
                if ( data.isHostingGame && data.currentGame.getPlayers().size() < MAX_PLAYERS) {
                    int sender = message.getSender();
                    announceGame(sender,HeartsMessage.HAVE_GAME);
                }
                break;

            case HeartsMessage.HAVE_GAME:
                Game game = (Game) m.getPayload();
                data.updateOpenGame(game);
                break;

            case HeartsMessage.JOIN_GAME_REQ:
                addRemotePlayer((Player) m.getPayload());
                break;

            case HeartsMessage.JOIN_GAME_ACK:
            	data.hasJoined = true;
            	data.notifyObservers();
            	break;
            
            case HeartsMessage.LEAVE_GAME:
                removeRemotePlayer((Player) m.getPayload());
                break;
                
            case HeartsMessage.ABORT_GAME:
            	data.hasJoined = false;
            	data.notifyObservers();         	
            	break;
            
            case HeartsMessage.START_GAME:
            	startGameThread((Game) m.getPayload());

        }
    }
}